# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
Yocto builder module
"""

import os.path
import shlex
from typing import List, Tuple, cast
from moulin.utils import create_stamp_name, construct_fetcher_dep_cmd
from moulin import ninja_syntax
from moulin.yaml_wrapper import YamlValue
from moulin.yaml_helpers import YAMLProcessingError


def get_builder(conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                generator: ninja_syntax.Writer):
    """
    Return configured YoctoBuilder class
    """
    return YoctoBuilder(conf, name, build_dir, src_stamps, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """
    Generate yocto build rules for ninja
    """
    # Create build dir by calling poky/oe-init-build-env script
    cmd = " && ".join([
        "cd $yocto_dir",
        "source poky/oe-init-build-env $work_dir",
    ])
    generator.rule("yocto_init_env",
                   command=f'bash -c "{cmd}"',
                   description="Initialize Yocto build environment",
                   restat=True)
    generator.newline()

    # Add bitbake layers by calling bitbake-layers script
    cmd = " && ".join([
        "cd $yocto_dir",
        "source poky/oe-init-build-env $work_dir",
        "bitbake-layers add-layer $layers",
        "touch $out",
    ])
    generator.rule("yocto_add_layers",
                   command=f'bash -c "{cmd}"',
                   description="Add yocto layers",
                   pool="console",
                   restat=True)
    generator.newline()

    # Write own configuration to moulin.conf. Include it in local.conf
    cmd = " && ".join([
        "cd $yocto_dir",
        "echo '# Code generated by moulin. All manual changes will be lost' > $work_dir/conf/moulin.conf",
        "for x in $conf; do echo $$x >> $work_dir/conf/moulin.conf; done",
        "sed \"/require moulin\\.conf/d\" -i $work_dir/conf/local.conf",
        "echo 'require moulin.conf' >> $work_dir/conf/local.conf",
        "touch -r $work_dir/conf/bblayers.conf $work_dir/conf/local.conf",
    ])
    generator.rule(
        "yocto_update_conf",
        command=cmd,
        description="Update local.conf",
    )
    generator.newline()

    # Invoke bitbake. This rule uses "console" pool so we can see the bitbake output.
    cmd = " && ".join([
        # Generate fetcher dependency file
        construct_fetcher_dep_cmd(),
        "cd $yocto_dir",
        "source poky/oe-init-build-env $work_dir",
        "bitbake $target",
    ])
    generator.rule("yocto_build",
                   command=f'bash -c "{cmd}"',
                   description="Yocto Build: $name",
                   pool="console",
                   deps="gcc",
                   depfile=".moulin_$name.d",
                   restat=True)


def _flatten_yocto_conf(conf: YamlValue) -> List[Tuple[str, str]]:
    """
    Flatten conf entries. While using YAML *entries syntax, we will get list of conf
    entries inside of other list. To overcome this, we need to move inner list 'up'
    """

    # Problem is conf entries that it is list itself
    result: List[Tuple[str, str]] = []
    for entry in conf:
        if not entry.is_list:
            raise YAMLProcessingError("Exptected array on 'conf' node", entry.mark)
        if entry[0].is_list:
            result.extend([(x[0].as_str, x[1].as_str) for x in entry])
        else:
            result.append((entry[0].as_str, entry[1].as_str))
    return result


class YoctoBuilder:
    """
    YoctoBuilder class generates Ninja rules for given build configuration
    """
    def __init__(self, conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                 generator: ninja_syntax.Writer):
        self.conf = conf
        self.name = name
        self.generator = generator
        self.src_stamps = src_stamps
        # With yocto builder it is possible to have multiple builds with the same set of
        # layers. Thus, we have two variables - build_dir and work_dir
        # - yocto_dir is the upper directory where layers are stored. Basically, we should
        #   have "poky" in our yocto_dir
        # - work_dir is the build directory where we can find conf/local.conf, tmp and other
        #   directories. It is called "build" by default
        self.yocto_dir = build_dir
        self.work_dir: str = conf.get("work_dir", "build").as_str

    def _get_external_src(self) -> List[Tuple[str, str]]:
        external_src_node = self.conf.get("external_src", None)
        if not external_src_node:
            return []

        ret: List[Tuple[str, str]] = []
        for key, val_node in cast(YamlValue, external_src_node).items():
            if val_node.is_list:
                path = os.path.join(*[x.as_str for x in val_node])
            else:
                path = val_node.as_str
            path = os.path.abspath(path)
            ret.append((f"EXTERNALSRC:pn-{key}", path))

        return ret

    def gen_build(self):
        """Generate ninja rules to build yocto/poky"""
        common_variables = {
            "yocto_dir": self.yocto_dir,
            "work_dir": self.work_dir,
        }

        # First we need to ensure that "conf" dir exists
        env_target = os.path.join(self.yocto_dir, self.work_dir, "conf", "local.conf")
        self.generator.build(env_target,
                             "yocto_init_env",
                             self.src_stamps,
                             variables=common_variables)

        # Then we need to add layers
        layers_node = self.conf.get("layers", None)
        if layers_node:
            layers_stamp = create_stamp_name(self.yocto_dir, self.work_dir, "yocto", "layers")
            layers = " ".join([x.as_str for x in layers_node])
            self.generator.build(layers_stamp,
                                 "yocto_add_layers",
                                 env_target,
                                 variables=dict(common_variables, layers=layers))

        # Next - update local.conf
        local_conf_target = os.path.join(self.yocto_dir, self.work_dir, "conf", "moulin.conf")
        local_conf_node = self.conf.get("conf", None)
        if local_conf_node:
            local_conf = _flatten_yocto_conf(local_conf_node)
        else:
            local_conf = []

        # Handle external sources (like build artifacts from some other build)
        local_conf.extend(self._get_external_src())

        # '$' is a ninja escape character so we need to quote it
        local_conf_lines = [
            shlex.quote(f'{k.replace("$", "$$")} = "{v.replace("$", "$$")}"') for k, v in local_conf
        ]

        self.generator.build(local_conf_target,
                             "yocto_update_conf",
                             layers_stamp if layers_node else env_target,
                             variables=dict(common_variables, conf=" ".join(local_conf_lines)))
        self.generator.newline()

        self.generator.build(f"conf-{self.name}", "phony", local_conf_target)
        self.generator.newline()

        # Next step - invoke bitbake. At last :)
        targets = self.get_targets()
        additional_deps_node = self.conf.get("additional_deps", None)
        if additional_deps_node:
            deps = [os.path.join(self.yocto_dir, d.as_str) for d in additional_deps_node]
        else:
            deps = []
        deps.append(local_conf_target)
        self.generator.build(targets,
                             "yocto_build",
                             deps,
                             variables=dict(common_variables,
                                            target=self.conf["build_target"].as_str,
                                            name=self.name))

        return targets

    def get_targets(self):
        "Return list of targets that are generated by this build"
        return [
            os.path.join(self.yocto_dir, self.work_dir, t.as_str)
            for t in self.conf["target_images"]
        ]

    def capture_state(self):
        """
        Update stored local conf with actual SRCREVs for VCS-based recipes.
        This should ensure that we can reproduce this exact build later
        """



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
Very minimal python module for generating Makefiles.
Loosely based on Ninja generator.
"""

import textwrap
from typing import List, Union, IO
from .ninja_syntax import as_list


class Writer:
    "Minimal Makefile Writter"

    def __init__(self, output: IO, width: int = 78):
        self.output = output
        self.width = width

    def _line(self, text: str) -> None:
        for line in textwrap.wrap(text,
                                  break_long_words=False,
                                  break_on_hyphens=False,
                                  subsequent_indent='  '):
            self.output.write(line + " \\\n")
        self.newline()

    def newline(self) -> None:
        "Emit a new line"
        self.output.write('\n')

    def comment(self, text) -> None:
        "Emit a comment"
        for line in textwrap.wrap(text,
                                  self.width - 2,
                                  break_long_words=False,
                                  break_on_hyphens=False):
            self.output.write('# ' + line + '\n')

    def simple_dep(self, outputs: Union[str, List[str]], inputs: Union[str, List[str]]) -> None:
        "Emit a simple dependency without build rules"
        outputs = as_list(outputs)
        inputs = as_list(inputs)

        self._line('%s: %s' % (' '.join(outputs), ' '.join(inputs)))

    def close(self) -> None:
        "Close the output file"
        self.output.close()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
sfdisk interface/wrapper for rouge image builder
"""

from pprint import pformat
from typing import List, Tuple, BinaryIO, Any
import subprocess
import logging

log = logging.getLogger(__name__)

DEFAULT_ALIGNMENT = 1 * 1024 * 1024  # 1 MiB


def _div_up(num: int, dem: int) -> int:
    if num % dem:
        return (num // dem) + 1
    return num // dem


def _sect(val, sector_size) -> int:
    return _div_up(val, sector_size)


def _align(val, align) -> int:
    return _div_up(val, align) * align


def _to_script(part: Any, sector_size=512) -> str:
    "Convert GPT Partition object to sfdisk script line"

    args = [
        f"start={_sect(part.start, sector_size)}", f"size={_sect(part.size, sector_size)}",
        f"type={part.gpt_type}", f"name={part.label}"
    ]

    if part.gpt_guid:
        args.append(f"uuid={part.gpt_guid}")

    return ", ".join(args)


def _check_sfdisk():
    # We are checking result explicitely
    # pylint: disable=subprocess-run-check
    ret = subprocess.run(["which", "sfdisk"], stdout=subprocess.DEVNULL)
    if ret.returncode != 0:
        raise Exception("Please make sure that 'sfdisk' is installed")


def _sfdisk_header():
    return "\n".join(["label: gpt", "unit: sectors"])


def fixup_partition_table(partitions: List[Any], sector_size=512) -> Tuple[List[Any], int]:
    """
    Return fixed partition table so it can be really written to disk.
    Also return total size of partition.
    """
    start_offset = 0
    if _sect(partitions[0].start, sector_size) < 2048:
        start_offset = 2048
    end = start_offset * sector_size
    ret = []
    for part in partitions:
        start = _align(end, DEFAULT_ALIGNMENT)  # Align to 1 MB
        size = _align(part.size, sector_size)
        ret.append(part._replace(start=start, size=size))
        end = start + size

    log.debug("Partition table: %s", pformat(ret))
    # Account for GPT copy
    return ret, end + 16 * 1024 * 1024


def write(fileo: BinaryIO, partitions: List[Any]):
    "Write partitions to a file"
    _check_sfdisk()

    # Generate sfdisk script file
    script = _sfdisk_header() + "\n"
    script += "\n".join(map(_to_script, partitions))

    log.debug("sfdisk script:\n---\n%s\n---", script)
    log.info("Creating GPT partition in %s", fileo.name)
    subprocess.run(["sfdisk", fileo.name],
                   input=bytes(script, 'UTF-8'),
                   check=True,
                   stdout=subprocess.DEVNULL)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""Module hosting different utility functions"""

import os.path
import sys


def create_stamp_name(*args):
    """Generate stamp file name based on input keywords"""
    stamp = "-".join(args)
    path = os.path.join(".stamps", stamp.replace("-", "--").replace(os.sep, "-").replace(":", "-"))
    return os.path.abspath(path)


def construct_fetcher_dep_cmd() -> str:
    "Generate command line to generate fetcher dependency file"
    this_script = os.path.abspath(sys.argv[0])
    args = " ".join(sys.argv[1:])
    return f"{this_script} {args} --fetcherdep $name"



"""
Intergration-level test for MoulinConfiguration
"""

import unittest
import unittest.mock

import yaml

from packaging.version import Version

from moulin.yaml_helpers import YAMLProcessingError
from moulin.build_conf import MoulinConfiguration


def gen_config(doc: str) -> MoulinConfiguration:
    node = yaml.compose(doc)
    return MoulinConfiguration(node)


class TestMoulinConfigurationBasic(unittest.TestCase):
    def test_desc(self):
        doc = """
desc: "Test build"
        """
        conf = gen_config(doc)
        self.assertEqual(conf.desc, "Test build")

    def test_no_desc(self):
        doc = """
 nodesc: ""
        """
        with self.assertRaisesRegex(YAMLProcessingError, "desc key is mandatory"):
            gen_config(doc)

    def test_min_ver(self):
        doc = """
desc: "Test build"
min_ver: "0.3"
        """
        conf = gen_config(doc)
        self.assertEqual(conf.min_ver, Version("0.3"))

    def test_no_min_ver(self):
        doc = """
desc: "Test build"
        """
        conf = gen_config(doc)
        self.assertIsNone(conf.min_ver)


class TestMoulinConfigurationVariables(unittest.TestCase):
    def test_simple(self):
        doc = """
desc: "Test build"
variables:
        A: test
        B: "%{A}%{A}"
        C: "%{B}%{B}"
        D: "%{A}x%{A}"
test:
        - "%{A}"
        - "%{B}"
        - "%{C}"
        - "%{D}"
"""
        expected = ["test", "testtest", "testtesttesttest", "testxtest"]
        conf = gen_config(doc)
        conf.complete_init(None)
        root = conf.get_root()
        self.assertEqual([x.as_str for x in root["test"]], expected)

    def test_loop(self):
        doc = """
desc: "Test build"
variables:
        A: "%{B}"
        B: "%{A}"
"""
        conf = gen_config(doc)
        with self.assertRaisesRegex(Exception, "circular dependency"):
            conf.complete_init(None)

    def test_self_loop(self):
        doc = """
desc: "Test build"
variables:
        A: "%{A}"
"""
        conf = gen_config(doc)
        with self.assertRaisesRegex(YAMLProcessingError, "refers to self"):
            conf.complete_init(None)

    def test_no_var(self):
        doc = """
desc: "Test build"
test: "%{A}"
"""
        conf = gen_config(doc)
        with self.assertRaisesRegex(YAMLProcessingError, "unknown variable"):
            conf.complete_init(None)

    def test_escape(self):
        doc = """
desc: "Test build"
test:
  - "%"
  - "%%"
  - "%%{A}"
  - "%{A"
"""
        expected = ["%", "%%", "%%{A}", "%{A"]
        conf = gen_config(doc)
        conf.complete_init(None)
        root = conf.get_root()
        self.assertEqual([x.as_str for x in root["test"]], expected)


class TestMoulinConfigurationOptions(unittest.TestCase):
    def test_list_params(self):
        doc = """
desc: "Test build"
parameters:
  paramA:
    desc: "Parameter A"
  paramB:
    desc: "Parameter B"
"""
        expected = ["paramA", "paramB"]
        conf = gen_config(doc)
        params = conf.get_parameters()
        self.assertEqual(list(params.keys()), expected)

    def test_list_variants(self):
        doc = """
desc: "Test build"
parameters:
  paramA:
    desc: "Parameter A"
    variantA:
      default: false
    variantB:
      default: false
    variantC:
      default: false
"""
        expected = ["variantA", "variantB", "variantC"]
        conf = gen_config(doc)
        param = conf.get_parameters()["paramA"]
        self.assertEqual(list(param.variants.keys()), expected)

    def test_default_variant(self):
        doc = """
desc: "Test build"
val: X
parameters:
  paramA:
    desc: "Parameter A"
    variantA:
      default: true
    variantB:
      default: false
"""
        conf = gen_config(doc)
        param = conf.get_parameters()["paramA"]
        self.assertIsNotNone(param.default)

    def test_apply_variant(self):
        doc = """
desc: "Test build"
val: X
parameters:
  paramA:
    desc: "Parameter A"
    variantA:
      overrides:
        val: A
    variantB:
      overrides:
        val: B
"""
        conf = gen_config(doc)
        conf.complete_init({"paramA": "variantA"})
        root = conf.get_root()
        self.assertEqual(root["val"].as_str, "A")

        conf = gen_config(doc)
        conf.complete_init({"paramA": "variantB"})
        root = conf.get_root()
        self.assertEqual(root["val"].as_str, "B")



# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 EPAM Systems
"""Unpacker fetcher module"""

import os.path
import subprocess
from typing import List
from moulin.yaml_helpers import YAMLProcessingError
from moulin.yaml_wrapper import YamlValue
from moulin import ninja_syntax


def get_fetcher(conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
    """Construct and return RepoFetcher object"""
    return UnpackFetcher(conf, build_dir, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """Generate build rules using Ninja generator"""
    generator.rule("tar_unpack",
                   command="mkdir -p $out_dir && tar -m -C $out_dir -xf $in",
                   description="Unpack $in with tar")
    generator.rule("zip_unpack",
                   command="mkdir -p $out_dir && unzip -DD -n $in -d $out_dir",
                   description="Unpack $in with unzip")
    generator.newline()


class UnpackFetcher:
    """
    Unpak fetcher class. Provides methods to generate rules for
    unpacking various archives
    """

    def __init__(self, conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
        self.conf = conf
        self.build_dir = build_dir
        self.generator = generator
        self.fname = conf["file"].as_str
        dirname = conf.get("dir", default=".").as_str
        self.out_dir = os.path.join(build_dir, dirname)
        self.type = conf["archive_type"].as_str

        known_types = ["tar", "zip"]
        if self.type not in known_types:
            raise YAMLProcessingError(f"Unkown archive type: {self.type}",
                                      self.conf["archive_type"].mark)

        if not os.path.exists(self.fname):
            raise YAMLProcessingError(f"File \"{self.fname}\" does not exist",
                                      self.conf["file"].mark)

    def gen_fetch(self) -> List[str]:
        """Generate instructions to unpack archive"""
        rule_name = f"{self.type}_unpack"
        files = self.get_file_list()
        self.generator.build(files, rule_name, self.fname, variables={"out_dir": self.out_dir})

        return files

    def get_file_list(self) -> List[str]:
        "Get list of files in archive"

        commands = {
            "tar": ["tar", "--list", "-f", f"{self.fname}"],
            "zip": ["unzip", "-Z", "-2", f"{self.fname}"]
        }
        ret = subprocess.run(commands[self.type],
                             check=True,
                             stdout=subprocess.PIPE,
                             universal_newlines=True)
        return [os.path.join(self.out_dir, x) for x in ret.stdout.split("\n") if len(x) > 0]



# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 EPAM Systems
"""
Null build generator. Used mostly for testing
"""

from typing import List
from moulin.yaml_wrapper import YamlValue
from moulin import ninja_syntax


def get_builder(conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                generator: ninja_syntax.Writer):
    """
    Return configured Null builder
    """
    return NullBuilder(conf, name, build_dir, src_stamps, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """
    Generate yocto build rules for ninja
    """
    generator.newline()


class NullBuilder:
    """
    NullBuilder class generates no Ninja rules
    """
    def __init__(self, conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                 generator: ninja_syntax.Writer):
        pass

    def gen_build(self):
        """Generate no Ninja rules"""

        return []

    def get_targets(self):
        "Return list of targets that are generated by this build"
        return []

    def capture_state(self):
        """
        This method should capture state for reproducible builds.
        """



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""Defines Configuration class that holds the the whole configuration object"""

import logging
import re
from typing import List, Optional, Dict, Union, NamedTuple
from functools import partial
from packaging.version import Version

import yaml
from yaml.nodes import MappingNode, ScalarNode, SequenceNode, Node
from moulin.yaml_helpers import get_boolean_value, get_str_value, get_mandatory_str_value, \
    get_mapping_node, YAMLProcessingError

from moulin.yaml_wrapper import YamlValue

log = logging.getLogger(__name__)


class ParameterVariant:
    "Represents possible parameter value"

    # pylint: disable=too-few-public-methods

    def __init__(self, name: str, node: MappingNode):
        self.name = name
        self.default, _ = get_boolean_value(node, "default", default=False)
        self._overrides: Optional[MappingNode] = None
        overrides_node = get_mapping_node(node, "overrides")
        if overrides_node:
            self._overrides = overrides_node

    def apply_overrides(self, node: MappingNode):
        "Apply variant overrides to a mapping node"
        if self._overrides:
            self._override(node, self._overrides)

    @staticmethod
    def _override(remote_node: Node, local_node: Node):
        # pylint: disable=unidiomatic-typecheck
        if type(remote_node) != type(local_node):
            raise YAMLProcessingError(
                f"Incompatible types {type(remote_node)} and " +
                f" {type(local_node)} while trying to apply overrides", local_node.start_mark)
        if isinstance(remote_node, MappingNode):
            for lname, lval in local_node.value:
                for i, rentry in enumerate(remote_node.value):
                    if rentry[0].value == lname.value:
                        if isinstance(rentry[1], ScalarNode):
                            remote_node.value[i] = (lname, lval)
                        else:
                            ParameterVariant._override(rentry[1], lval)
                        break
                else:
                    remote_node.value.append((lname, lval))
        elif isinstance(remote_node, SequenceNode):
            remote_node.value.extend(local_node.value)
        else:
            raise YAMLProcessingError(f"Unknown node type f{type(remote_node)}",
                                      remote_node.start_mark)


class Parameter:
    "Represents parametrization option"

    # pylint: disable=too-few-public-methods

    def __init__(self, name: str, node: MappingNode):
        self.name = name
        self.desc, _ = get_mandatory_str_value(node, "desc")

        self.variants: Dict[str, ParameterVariant] = {}
        vname: Node
        vnode: Node
        for vname, vnode in node.value:
            if not isinstance(vname, ScalarNode) or not isinstance(vname.value, str):
                raise YAMLProcessingError("Variant name is expected to be string", vname.start_mark)
            if vname.value == "desc":
                continue
            if not isinstance(vnode, MappingNode):
                raise YAMLProcessingError("Variant definition is expected to be a dict",
                                          vnode.start_mark)
            self.variants[vname.value] = ParameterVariant(vname.value, vnode)

        self.default: Optional[ParameterVariant] = None
        for variant in self.variants.values():
            if variant.default:
                self.default = variant
                break

    def apply_overrides(self, node: MappingNode, variant_name: str):
        "Apply given variant to a mapping node"

        self.variants[variant_name].apply_overrides(node)


escaped_percent = re.compile(r"(%%)")
variable_re = re.compile(r"%\{(\w[\w\d\-]*)\}")


class VariableRef(NamedTuple):
    "Represents variable reference"
    name: str


class Variable(NamedTuple):
    "Represents variable"
    name: str
    tokens: List[Union[str, VariableRef]]
    mark: yaml.Mark


class ExpandedVariable(NamedTuple):
    "Represents expanded variable"
    name: str
    value: str
    mark: yaml.Mark


class MoulinConfiguration:
    "This class holds the whole build configuration"

    def __init__(self, node: MappingNode):
        self._node = node

        min_ver, _ = get_str_value(node, "min_ver")
        if not min_ver:
            self.min_ver = None
        else:
            self.min_ver = Version(min_ver)

        self.desc, _ = get_mandatory_str_value(node, "desc")

        parameters_node = get_mapping_node(node, "parameters")
        self._params: Dict[str, Parameter] = {}
        if parameters_node:
            self._read_params(parameters_node)

        self._variables: Dict[str, ExpandedVariable] = {}

    def _read_params(self, node: MappingNode):
        pname: Node
        pnode: Node
        for pname, pnode in node.value:
            if not isinstance(pname, ScalarNode) or not isinstance(pname.value, str):
                raise YAMLProcessingError("Parameter node is expected to be string",
                                          pname.start_mark)
            if not isinstance(pnode, MappingNode):
                raise YAMLProcessingError("Parameter variants is expected to be a dict",
                                          pnode.start_mark)
            self._params[pname.value] = Parameter(pname.value, pnode)

    def get_parameters(self):
        "Return available parameters"
        return self._params

    def _prepare_variables(self):
        variables_node = get_mapping_node(self._node, "variables")
        variables: Dict[str, Variable] = {}
        if not variables_node:
            return
        for vname_node, vval_node in variables_node.value:
            if not isinstance(vname_node, ScalarNode) or not isinstance(vname_node.value, str):
                raise YAMLProcessingError("Variable name is expected to be string",
                                          vname_node.start_mark)
            if not isinstance(vval_node, ScalarNode) or not isinstance(vval_node.value, str):
                raise YAMLProcessingError("Variable value is expected to be string",
                                          vval_node.start_mark)
            vname = vname_node.value
            variables[vname] = Variable(vname, _tokenize(vval_node.value), vname_node.start_mark)
            # Sanity check
            for token in variables[vname].tokens:
                if isinstance(token, VariableRef):
                    if token.name == vname:
                        raise YAMLProcessingError(f"Variable {vname} refers to self",
                                                  vname_node.start_mark)
        self._variables = _try_to_expand_variables(variables)

    def complete_init(self, options):
        "Complete object initialization with given options"
        if options:
            log.info("Completing setup with following parameters:")
            for param_name, param_value in options.items():
                log.info("  %s: %s", param_name, param_value)
                self._params[param_name].apply_overrides(self._node, param_value)

        log.debug("Expanding variables...")
        self._prepare_variables()
        for var in self._variables.values():
            log.debug("  %s = %s", var.name, var.value)

        log.debug("Substituting variables...")
        _traverse_tree(self._node, partial(_substitute_variables, variables=self._variables))

        log.debug("Cleaning up config...")
        # Drop 'variables' and 'parameters'
        self._node.value = list(
            filter(lambda x: x[0].value not in ["variables", "parameters"], self._node.value))

    def dumps(self):
        "Return processed YAML stream"
        return yaml.serialize(self._node)

    def get_root_node(self):
        "Return root YAML node"
        return self._node

    def get_root(self) -> YamlValue:
        "Return root as YamlValue object"
        return YamlValue(self._node)


#
# Variable processing helpers.
#


def _extract_refs(string: str) -> List[Union[str, VariableRef]]:
    ret: List[Union[str, VariableRef]] = []
    for i, part in enumerate(variable_re.split(string)):
        # Even (zero-based) entries are just strings
        # Odd entries are variable references
        if i % 2 == 0 and part:
            ret.append(part)
        elif part:
            ret.append(VariableRef(part))
    return ret


def _tokenize(string: str) -> List[Union[str, VariableRef]]:
    ret: List[Union[str, VariableRef]] = []
    for part in escaped_percent.split(string):
        if part:
            if escaped_percent.fullmatch(part):
                ret.append("%%")
            else:
                ret.extend(_extract_refs(part))
    return ret


def _map_token(token: Union[str, VariableRef], variables: Dict[str, ExpandedVariable],
               mark: yaml.Mark):
    if isinstance(token, str):
        return token
    if isinstance(token, VariableRef):
        name = token.name
        if name not in variables:
            return token
        return variables[name].value
    raise YAMLProcessingError(f"Unexptected token of type '{type(token)}'", mark)


def _contains_ref(tokens):
    for token in tokens:
        if isinstance(token, VariableRef):
            return True
    return False


def _list_compare(list1, list2):
    "Compare contents of lists, not list object themselves"
    if len(list1) != len(list2):
        return False
    for i, j in zip(list1, list2):
        if i != j:
            return False
    return True


def _try_to_expand_variables(variables: Dict[str, Variable]) -> Dict[str, ExpandedVariable]:
    # This is not most effective way. But it is good enough. Iterate
    # over all variables, trying to expand one of them
    expanded_variables: Dict[str, ExpandedVariable] = {}
    stop = False
    while not stop:
        stop = True
        for key, var in variables.items():
            new_tokens = list(
                map(partial(_map_token, variables=expanded_variables, mark=var.mark), var.tokens))
            if not _contains_ref(new_tokens):
                expanded_variables[key] = ExpandedVariable(var.name, "".join(new_tokens), var.mark)
                del variables[key]
                stop = False
                break
            if not _list_compare(var.tokens, new_tokens):
                variables[key] = Variable(var.name, new_tokens, var.mark)

    # We can't expand more variables. Either all are expanded, or
    # there is a some user error. Perform sanity check.
    for var in variables.values():
        for token in var.tokens:
            if isinstance(token, VariableRef):
                if token.name == var.name:
                    raise YAMLProcessingError(f"Variable {var.name} indirectly refers to self",
                                              var.mark)
                if token.name not in variables:
                    raise YAMLProcessingError(
                        f"Variable {var.name} refers to unknown variable {token.name}", var.mark)
    if variables:
        raise Exception(f"Found circular dependency in variables {list(variables.keys())}")

    return expanded_variables


def _substitute_variables(node: ScalarNode, variables: Dict[str, ExpandedVariable]):
    if not isinstance(node.value, str):
        return
    tokens = _tokenize(node.value)
    tokens = list(map(partial(_map_token, variables=variables, mark=node.start_mark), tokens))
    for token in tokens:
        if isinstance(token, VariableRef):
            raise YAMLProcessingError(f"Reference to unknown variable {token.name}",
                                      node.start_mark)
    node.value = "".join(tokens)  # type: ignore # there are only strings left in tokens


def _traverse_tree(node, fn):
    # pylint: disable=invalid-name
    if isinstance(node, MappingNode):
        for name, val in node.value:
            _traverse_tree(name, fn)
            _traverse_tree(val, fn)
    elif isinstance(node, SequenceNode):
        for subnode in node.value:
            _traverse_tree(subnode, fn)
    elif isinstance(node, ScalarNode):
        fn(node)
    else:
        raise YAMLProcessingError(f"Unknown node type {type(node)}", node.start_mark)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
Android kernel builder module
"""

import os.path
from typing import List
from moulin.yaml_wrapper import YamlValue
from moulin import ninja_syntax


def get_builder(conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                generator: ninja_syntax.Writer):
    """
    Return configured AndroidKernel class
    """
    return AndroidKernel(conf, name, build_dir, src_stamps, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """
    Generate yocto build rules for ninja
    """
    cmd = " && ".join([
        "export $env",
        "cd $build_dir",
        "build/build.sh",
    ])
    generator.rule("android_kernel_build",
                   command=f'bash -c "{cmd}"',
                   description="Invoke Android Kernel build script",
                   pool="console")
    generator.newline()


class AndroidKernel:
    """
    AndroidBuilder class generates Ninja rules for given Android build configuration
    """
    def __init__(self, conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                 generator: ninja_syntax.Writer):
        self.conf = conf
        self.name = name
        self.generator = generator
        self.src_stamps = src_stamps
        self.build_dir = build_dir

    def gen_build(self):
        """Generate ninja rules to build AOSP"""

        env_node = self.conf.get("env", None)
        if env_node:
            env_values = [x.as_str for x in env_node]
        else:
            env_values = []
        env = " ".join(env_values)
        variables = {
            "build_dir": self.build_dir,
            "env": env,
        }
        targets = self.get_targets()
        self.generator.build(targets, "android_kernel_build", self.src_stamps, variables=variables)
        self.generator.newline()

        return targets

    def get_targets(self):
        "Return list of targets that are generated by this build"
        return [os.path.join(self.build_dir, t.as_str) for t in self.conf["target_images"]]

    def capture_state(self):
        """
        This method should capture Android Kernel state for a reproducible builds.
        Luckily, there is nothing to do, as Android state is controlled solely by
        its repo state. And repo state is captured by repo fetcher code.
        """



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""YAML processing helper functions and objects"""

from typing import Optional, TypeVar, Tuple, List, Any, cast
from yaml.nodes import MappingNode, ScalarNode, SequenceNode, Node
from yaml.constructor import SafeConstructor
from yaml import Mark


class YAMLProcessingError(Exception):
    "Error during processing YAML file"

    def __init__(self, message: str, mark: Mark):
        super().__init__()
        self.message = message
        self.mark = mark

    def __str__(self):
        return f"{self.message} {self.mark}"


def get_node(node: MappingNode, name: str) -> Optional[Node]:
    "Return node with given name from mapping"
    if not isinstance(node, MappingNode):
        raise YAMLProcessingError("Not a mapping node", node.start_mark)
    for subpair in node.value:
        if subpair[0].value == name:
            return subpair[1]
    return None


def get_scalar_node(node: MappingNode, name: str) -> Optional[ScalarNode]:
    "Return scalar node with given name from mapping"
    value = get_node(node, name)
    if not value:
        return None
    if not isinstance(value, ScalarNode):
        raise YAMLProcessingError("Expected scalar value", value.start_mark)
    return value


def get_mapping_node(node: MappingNode, name: str) -> Optional[MappingNode]:
    "Return mapping node with given name from mapping"
    value = get_node(node, name)
    if not value:
        return None
    if not isinstance(value, MappingNode):
        raise YAMLProcessingError("Expected mapping", value.start_mark)
    return value


def get_sequence_node(node: MappingNode, name: str) -> Optional[SequenceNode]:
    "Return sequence node with given name from mapping"
    value = get_node(node, name)
    if not value:
        return None
    if not isinstance(value, SequenceNode):
        raise YAMLProcessingError("Expected sequence", value.start_mark)
    return value


def get_mandatory_mapping_node(node: MappingNode, name: str) -> MappingNode:
    "Return mandatory mapping node with given name from mapping"
    ret = get_mapping_node(node, name)
    if not ret:
        raise YAMLProcessingError(f"Key '{name} is mandatory'", node.start_mark)
    return ret


def get_mandatory_sequence_node(node: MappingNode, name: str) -> SequenceNode:
    "Return mandatory sequence node with given name from mapping"
    ret = get_sequence_node(node, name)
    if not ret:
        raise YAMLProcessingError(f"Key '{name} is mandatory'", node.start_mark)
    return ret


def flatten_list(node: SequenceNode) -> None:
    "Flatten sequence of sequences in-place"

    # Copy contents of node.value into a new list
    # because we will be modifying original one
    for element in list(node.value):
        if isinstance(element, SequenceNode):
            node.value.remove(element)
            node.value.extend(element.value)


T = TypeVar('T')  # pylint: disable=invalid-name

yaml_constructor = SafeConstructor()


def get_typed_value(mapping_node: MappingNode,
                    name: str,
                    expected_type: type,
                    default: Optional[T] = None) -> Tuple[Optional[T], Optional[Mark]]:
    "Return scalar node with given name from mapping as value of type T"
    node = get_scalar_node(mapping_node, name)
    if not node:
        return default, None
    val = yaml_constructor.construct_object(node)
    if not isinstance(val, expected_type):
        raise YAMLProcessingError(
            f"Expected value of type {expected_type.__name__} for " +
            f" '{name}', not {type(val).__name__}", node.start_mark)
    return cast(Optional[T], val), node.start_mark


def get_mandatory_typed_value(mapping_node: MappingNode, name: str,
                              expected_type: type) -> Tuple[T, Mark]:
    "Return mandatory scalar node with given name from mapping as value of type T"
    ret, mark = get_typed_value(mapping_node, name, expected_type)
    if not ret or not mark:
        raise YAMLProcessingError(f"{name} key is mandatory", mapping_node.start_mark)
    return ret, mark


def get_boolean_value(mapping_node: MappingNode,
                      name: str,
                      default: Optional[bool] = None) -> Tuple[Optional[bool], Optional[Mark]]:
    "Return scalar node with given name from mapping as a boolean value"
    return get_typed_value(mapping_node, name, bool, default)


def get_str_value(mapping_node: MappingNode,
                  name: str,
                  default: Optional[str] = None) -> Tuple[Optional[str], Optional[Mark]]:
    "Return scalar node with given name from mapping as a string value"
    return get_typed_value(mapping_node, name, str, default)


def get_mandatory_str_value(mapping_node: MappingNode, name: str) -> Tuple[str, Mark]:
    "Return scalar node with given name from mapping as a string value"
    return get_mandatory_typed_value(mapping_node, name, str)


def get_mandatory_mapping(node: MappingNode, name: str) -> List[Tuple[Node, Any]]:
    "Return mandatory sequence node with given name from mapping"
    return get_mandatory_mapping_node(node, name).value



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
External utils interfaces/wrappers for rouge image builder
"""

from typing import BinaryIO, Union, Optional
import subprocess
import logging

log = logging.getLogger(__name__)


def _run_cmd(args):
    log.info("Running %s", " ".join(args))
    subprocess.run(args, check=True)


# pylint: disable=invalid-name
def dd(file_in: Union[str, BinaryIO], file_out: BinaryIO, out_offset: int, out_size: Optional[int] = None):
    "Run dd with the given arguments"
    # Try to guess block size. We would like to use as big block as
    # possible. But we need take into account that "seek" parameter
    # uses block size as the unit.
    blocksize: int = 65536
    while out_offset % blocksize != 0:
        blocksize //= 2

    if isinstance(file_in, str):
        file_in_path = file_in
    else:
        file_in_path = file_in.name
    args = [
        "dd",
        f"if={file_in_path}",
        f"of={file_out.name}",
        f"bs={blocksize}",
        f"seek={out_offset // blocksize}",
        "status=progress",
        "conv=sparse",
        "conv=notrunc",
    ]  # yapf: disable
    if out_size:
        args.append(f"count={out_size // blocksize}")
    _run_cmd(args)


def simg2img(file_in: Union[str, BinaryIO], file_out: BinaryIO):
    "Run simg2img with the given arguments"
    if isinstance(file_in, str):
        file_in_path = file_in
    else:
        file_in_path = file_in.name
    args = [
        "simg2img",
        file_in_path,
        file_out.name,
    ]  # yapf: disable
    _run_cmd(args)


def mkext4fs(file_out: BinaryIO, contents_dir=None):
    "Create ext4 fs in given file"
    args = ["mkfs.ext4", file_out.name]
    if contents_dir:
        args.append("-d")
        args.append(contents_dir)

    _run_cmd(args)


def mkvfatfs(file_out: BinaryIO):
    "Create ext4 fs in given file"
    args = ["mkfs.vfat", file_out.name]

    _run_cmd(args)


def mcopy(img: BinaryIO, file: str, name: str):
    "Copy a file to a vfat image with a given name"
    args = ["mcopy", "-i", img.name, file, "::" + name]

    _run_cmd(args)


def mmd(img: BinaryIO, folders: list):
    "Create directories inside a vfat image"
    args = ["mmd", "-i", img.name]
    args.extend(folders)

    _run_cmd(args)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 EPAM Systems
"""West fetcher module"""

import shlex
import os.path
import subprocess
import pygit2
from typing import List
from moulin.utils import create_stamp_name
from moulin.yaml_wrapper import YamlValue
from moulin import ninja_syntax


def get_fetcher(conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
    """Construct and return WestFetcher object"""
    return WestFetcher(conf, build_dir, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """Generate build rules using Ninja generator"""

    cmd = " && ".join([
        "mkdir -p $build_dir",
        "cd $build_dir",
        "west init $west_args",
    ])
    generator.rule("west_init",
                   command=cmd,
                   description="Creating a west workspace (west init)")
    generator.newline()

    # `west update -n` is used here to narrow down fetched data
    # and significantly reduce number of output lines.
    # See `west help update` for additional info.
    cmd = " && ".join([
        "cd $build_dir",
        "west update -n",
        "touch $out",
    ])
    generator.rule("west_update",
                   command=cmd,
                   description="west update")
    generator.newline()


class WestFetcher:
    """
    West fetcher class. Provides methods to generate rules for
    fetching zephyr-based repositories using west meta-tool
    """
    def __init__(self, conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
        self.conf = conf
        self.build_dir = build_dir
        self.generator = generator

    def gen_fetch(self):
        """Generate instructions to fetch west-based repository"""
        west_args = []

        url = self.conf.get("url", "").as_str
        if url:
            west_args.append(f"-m {shlex.quote(url)}")
        rev = self.conf.get("rev", "").as_str
        if rev:
            west_args.append(f"--mr {shlex.quote(rev)}")
        filename = self.conf.get("file", "").as_str
        if filename:
            west_args.append(f"--mf {shlex.quote(filename)}")

        init_target = os.path.join(self.build_dir, ".west")
        update_target = create_stamp_name(self.build_dir, "update")

        self.generator.build(init_target,
                             "west_init",
                             variables={
                                 "build_dir": self.build_dir,
                                 "west_args": " ".join(west_args)
                             })
        self.generator.newline()

        self.generator.build(update_target,
                             "west_update",
                             init_target,
                             variables={
                                 "build_dir": self.build_dir
                             })
        self.generator.newline()

        return update_target

    def get_file_list(self) -> List[str]:
        """Get list of files under version control"""

        # First, get list of projects
        west_out = subprocess.run(["west", "list", "--format='{path}'"],
                                  check=True,
                                  cwd=self.build_dir,
                                  stdout=subprocess.PIPE,
                                  encoding="utf-8")
        projects = [x.strip("'") for x in west_out.stdout.split("\n")]

        # Then interrogate each git repo
        result: List[str] = []
        for project in projects:
            if len(project) == 0:
                continue
            git_dir = os.path.join(self.build_dir, project)
            repo = pygit2.Repository(git_dir)
            index = repo.index
            index.read()
            for entry in index:
                path = os.path.join(git_dir, entry.path)
                if os.path.isfile(path):
                    result.append(path.replace("$", "$$"))
        return result

    def capture_state(self):
        """Update provided conf with the current fetcher state"""
        raise Exception("capture_state is not implemented for west fetcher")



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
Moulin main entry point
"""

import argparse
import logging
import sys
from time import time
from datetime import timedelta
from pathlib import Path
from typing import Dict, List, Tuple, Any
import importlib_metadata
import yaml
from urllib.parse import urlparse, unquote
import urllib.request
import os.path

from packaging.version import Version

from moulin import build_generator
from moulin.build_conf import MoulinConfiguration
import moulin.rouge
import moulin.rouge.block_entry

log = logging.getLogger(__name__)

OptionDef = Tuple[List[str], Dict[str, Any]]


def _prepre_shared_opts(description: str,
                        additional_opts: List[OptionDef] = None,
                        exclusive_opts: List[List[OptionDef]] = None):

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('conf',
                        metavar='build.yaml',
                        type=str,
                        help='YAML file with build description')
    parser.add_argument('--help-config',
                        action='store_true',
                        help="Show help for given config file")
    parser.add_argument("-v", "--verbose", action="store_true", help="Enable verbose output")
    parser.add_argument("--dump", action="store_true", help="Dump processed YAML document")

    if additional_opts:
        for option_args, option_kwargs in additional_opts:
            parser.add_argument(*option_args, **option_kwargs)
    if exclusive_opts:
        for exclusive_set in exclusive_opts:
            group = parser.add_mutually_exclusive_group()
            for option_args, option_kwargs in exclusive_set:
                group.add_argument(*option_args, **option_kwargs)

    return parser


def _get_conf_file(conf_location: str) -> str:
    """Try to determine config file location. Check if input string is URL
    and download it if output file is not already present. If input string is
    not an URL, then assume that this is a local file and return it as is."""
    local_filename = conf_location

    parsed_url = urlparse(conf_location)
    if parsed_url.scheme:
        # it's not local file, so we extract filename
        # and download the file if it isn't present
        local_filename = os.path.basename(unquote(parsed_url.path))
        if os.path.exists(local_filename):
            # preserve possible local changes
            log.warn("Local copy already exists. Nothing has been downloaded.")
        else:
            log.info("Downloading from URL...")
            urllib.request.urlretrieve(conf_location, local_filename)

    return local_filename


def _handle_shared_opts(description: str,
                        additional_opts: List[OptionDef] = None,
                        exclusive_opts: List[List[OptionDef]] = None):

    parser = _prepre_shared_opts(description, additional_opts, exclusive_opts)

    args, extra_opts = parser.parse_known_args()

    loglevel = logging.INFO
    if args.verbose:
        loglevel = logging.DEBUG
    logging.basicConfig(level=loglevel, format="[%(levelname)s] %(message)s")

    local_conf_file = _get_conf_file(args.conf)

    conf = MoulinConfiguration(yaml.compose(open(local_conf_file)))

    if conf.min_ver:
        our_ver = Version(importlib_metadata.version("moulin"))
        if our_ver < conf.min_ver:
            raise Exception(f"Config file requires version {conf.min_ver}," +
                            f" while you are running moulin {our_ver}")

    prog = f"{sys.argv[0]} {local_conf_file}"
    desc = f"Config file description: {conf.desc}"
    config_argparser = argparse.ArgumentParser(description=desc, prog=prog, add_help=False)
    for parameter in conf.get_parameters().values():
        config_argparser.add_argument(f"--{parameter.name}",
                                      choices=[x.name for x in parameter.variants.values()],
                                      default=parameter.default.name,
                                      help=parameter.desc)

    if args.help_config:
        config_argparser.print_help()
        sys.exit(0)

    config_args = config_argparser.parse_args(extra_opts)
    conf.complete_init(vars(config_args))

    if args.dump:
        print(conf.dumps())

    return conf, args


def moulin_entry():
    """Console entry point for moulin"""

    additional_opts = [
        (["--fetcherdep"], dict(nargs=1, metavar="component", help=argparse.SUPPRESS)),
    ]
    conf, args = _handle_shared_opts(
        f'Moulin meta-build system v{Version(importlib_metadata.version("moulin"))}',
        additional_opts=additional_opts)
    log.info("Generating build.ninja")

    if not args.fetcherdep:
        build_generator.generate_build(conf, args.conf)
    else:
        log.info("Generating deps for component '%s'", args.fetcherdep[0])
        build_generator.generate_fetcher_dyndep(conf, args.fetcherdep[0])


def rouge_entry():
    """Console entry point for rouge"""

    start_time = time()

    additional_opts = [
        (["-f", "--force"], dict(action="store_true", help="Force overwrite output file")),
        (["-s", "--special"],
         dict(action="store_true", help="Allow writing to special files (like block devices)")),
        (["-o"],
         dict(nargs=1,
              metavar="FILE",
              dest="output",
              help="Output file name (defaults to <image_name>.img")),
    ]
    exclusive_opts = [
        [
            (["-l", "--list-images"], dict(action="store_true",
                                           help="List images defined in config")),
            (["-i", "--image-name"],
             dict(nargs=1,
                  metavar="image_name",
                  help="Name of image (which is defined in config file)")),
        ],
    ]

    conf, args = _handle_shared_opts(
        f'Rouge boot image generator v{Version(importlib_metadata.version("moulin"))}',
        additional_opts, exclusive_opts)
    root = conf.get_root()
    images = moulin.rouge.get_available_images(root)

    if not images:
        log.error("No images defined in the provided build description file")
        sys.exit(1)

    if args.list_images:
        print("Listing available images")
        for image in images:
            print(f" - {image.name}: {image.desc}")
        sys.exit(0)

    if not args.image_name:
        log.error("Either -l or -i parameter is required")
        sys.exit(1)

    output_file: str
    if args.output:
        output_file = args.output[0]
    else:
        output_file = f"{args.image_name[0]}.img"

    # Okay, time to really write something
    for image in images:
        if image.name == args.image_name[0]:
            chosen_image = image
            break
    else:
        log.error("Can't find requested image. Use '-l' option to see available images.")
        sys.exit(1)

    output_path = _rouge_validate_output(output_file, args)

    block_entry = moulin.rouge.block_entry.construct_entry(chosen_image.node)
    with open(output_file, "wb") as fileo:
        if not output_path.is_block_device():
            fileo.truncate(block_entry.size())
        block_entry.write(fileo, 0)

    end_time = time()
    log.info("Done in %s", timedelta(seconds=(end_time - start_time)))

    # Good advice
    if output_path.is_file():
        log.info(" ".join([
            "You can write result to your SD card.",
            "Don't forget to pass 'conv=sparse' option to dd.",
            "This will improve writing speed greatly."
        ]))


def _rouge_validate_output(output_file: str, args):
    output_path = Path(output_file)
    if output_path.is_file() and not args.force:
        log.error("Output file %s already exists and no '-f' flag provided. Exiting.", output_file)
        sys.exit(1)

    if output_path.is_block_device() and not args.special:
        log.error("Output file %s is a block device and no '-s' flag provided. Exiting.",
                  output_file)
        sys.exit(1)
    else:
        if output_path.exists() and not output_path.is_file() and not output_path.is_block_device():
            log.error("Output path %s exists and it is not a file or a block device. Exiting.",
                      output_file)
            sys.exit(1)

    if output_path.is_block_device():
        log.warning("Writing directly to block device %s. We hope you provided correct one.",
                    output_file)

    if output_path.is_file():
        log.warning("Overwriting file %s", output_file)

    return output_path



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
Wrappers for yaml.Node that provide better API
"""

from typing import Optional, List, Tuple, Iterator, Union

from yaml.nodes import MappingNode, ScalarNode, SequenceNode, Node
from yaml.constructor import SafeConstructor
from yaml.representer import SafeRepresenter
from yaml import Mark
from moulin.yaml_helpers import YAMLProcessingError

yaml_constructor = SafeConstructor()


class _YamlDefaultValue:
    """
    Helper class that have the same API as YamlValue, but is
    constructed from a primitive type. It is used to provide default
    value in YamlValue.get() method
    """
    def __init__(self, val: Union[bool, str, int, float, None]):
        self._val = val

    def __bool__(self):
        return self._val is not None

    @property
    def as_bool(self) -> bool:
        "Get the boolean value"
        if not isinstance(self._val, bool):
            raise TypeError("Expected boolean value")
        return self._val

    @property
    def as_str(self) -> str:
        "Get the string value"
        if not isinstance(self._val, str):
            raise TypeError("Expected string value")
        return self._val

    @property
    def as_int(self) -> int:
        "Get the integer value"
        if not isinstance(self._val, int):
            raise TypeError("Expected int value")
        return self._val

    @property
    def as_float(self) -> float:
        "Get the floating point value"
        if not isinstance(self._val, int):
            raise TypeError("Expected float value")
        return self._val


class YamlValue:  # pylint: disable=too-few-public-methods
    """Wrapper for yaml.Node class. It provides type-safe access to YAML nodes"""
    def __init__(self, node: Node):
        self._node = node
        if isinstance(node, ScalarNode):
            self._val = yaml_constructor.construct_object(node)
        else:
            self._val = None

    @property
    def mark(self) -> Mark:
        "Get the start mark for YAML Node"
        return self._node.start_mark

    @property
    def as_bool(self) -> bool:
        "Get the boolean value"
        if not isinstance(self._val, bool):
            raise YAMLProcessingError(f"Expected boolean value, got {type(self._val)}", self.mark)
        return self._val

    @property
    def as_str(self) -> str:
        "Get the string value"
        if not isinstance(self._val, str):
            raise YAMLProcessingError(f"Expected string value, got {type(self._val)}", self.mark)
        return self._val

    @property
    def as_int(self) -> int:
        "Get the integer value"
        if not isinstance(self._val, int):
            raise YAMLProcessingError(f"Expected integer value, got {type(self._val)}", self.mark)
        return self._val

    @property
    def as_float(self) -> float:
        "Get the floating point value"
        if not isinstance(self._val, float):
            raise YAMLProcessingError(f"Expected floating point value, got {type(self._val)}",
                                      self.mark)
        return self._val

    @property
    def is_list(self) -> bool:
        """Check if this node represents a list"""
        return isinstance(self._node, SequenceNode)

    def _get(self, name: str) -> Optional["YamlValue"]:
        "Get optional value by name"
        if not isinstance(self._node, MappingNode):
            raise YAMLProcessingError("Mapping node is expected", self.mark)
        for key, val in self._node.value:
            if key.value == name:
                return YamlValue(val)
        return None

    def get(self, name: str, default) -> Union["YamlValue", _YamlDefaultValue]:
        "Get optional value by name"
        val = self._get(name)
        if val:
            return val
        return _YamlDefaultValue(default)

    def keys(self) -> List[str]:
        """Get all keys for this mapping"""
        if not isinstance(self._node, MappingNode):
            raise YAMLProcessingError("Mapping node is expected", self.mark)
        return [key.value for key, _ in self._node.value]

    def items(self) -> List[Tuple[str, "YamlValue"]]:
        """Get all items for this mapping"""
        if not isinstance(self._node, MappingNode):
            raise YAMLProcessingError("Mapping node is expected", self.mark)
        return [(key.value, YamlValue(val)) for key, val in self._node.value]

    def replace_value(self, val: Union[str, int, bool, float]):
        "Set a new value for a scalar node"
        if not isinstance(self._node, ScalarNode):
            raise YAMLProcessingError("Can't replace value for a non-scalar node", self.mark)
        self._node.value = val

    def __getitem__(self, idx: Union[str, int]) -> "YamlValue":
        if isinstance(idx, int):
            if not isinstance(self._node, SequenceNode):
                raise YAMLProcessingError("SequenceNode node is expected", self.mark)
            return YamlValue(self._node.value[idx])
        if isinstance(idx, str):
            val = self._get(idx)
            if not val:
                raise YAMLProcessingError(f"Key '{idx}' is mandatory", self.mark)
            return val
        raise KeyError("Key should have either type 'str' or 'int'")

    def __setitem__(self, idx: Union[str, int], val: Union[str, int, bool, float]):
        if isinstance(idx, int):
            if not isinstance(self._node, SequenceNode):
                raise YAMLProcessingError("SequenceNode node is expected", self.mark)
            self._node.value[idx].replace_value(val)
        if isinstance(idx, str):
            item = self._get(idx)
            if item:
                item.replace_value(val)
            else:
                representer = SafeRepresenter()
                key_node = representer.represent_str(idx)
                if isinstance(val, str):
                    val_node = representer.represent_str(val)
                elif isinstance(val, int):
                    val_node = representer.represent_int(val)
                elif isinstance(val, bool):
                    val_node = representer.represent_bool(val)
                else:
                    val_node = representer.represent_float(val)
                self._node.value.append((key_node, val_node))
        raise KeyError("Key should have either type 'str' or 'int'")

    def __iter__(self) -> Iterator["YamlValue"]:
        for item in self._node.value:
            yield YamlValue(item)

    def __len__(self) -> int:
        return len(self._node.value)

    def __contains__(self, key: str) -> bool:
        """Test if key is present in mapping"""
        if not isinstance(self._node, MappingNode):
            raise YAMLProcessingError("Mapping node is expected", self.mark)
        return key in self.keys()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
Main 'rouge' logic lives there
"""

import os.path
import struct
import shutil
import logging
import itertools
from typing import List, Tuple, NamedTuple, cast
from tempfile import NamedTemporaryFile, TemporaryDirectory

from typing import Optional
from yaml import Mark
from moulin.rouge import sfdisk, ext_utils
from moulin.yaml_helpers import YAMLProcessingError
from moulin.yaml_wrapper import YamlValue

log = logging.getLogger(__name__)


class BlockEntry():
    "Base class for various block entries"

    # pylint: disable=too-few-public-methods

    def write(self, _file, _offset):
        "write() in base class does nothing"

    def get_deps(self) -> List[str]:  # pylint: disable=no-self-use
        "get_deps() in base class does nothing"
        return []


class GPTPartition(NamedTuple):
    "Represents one partition in GPT"
    label: str
    gpt_type: str
    gpt_guid: str
    start: int
    size: int
    entry: BlockEntry


class GPT(BlockEntry):
    "Represents GUID Partition Table"

    def __init__(self, node: YamlValue):
        self._partitions: List[GPTPartition] = []
        self._size: int = 0
        self._requested_image_size: Optional[int] = None

        _requested_image_size_node = node.get("image_size", None)
        if _requested_image_size_node:
            self._requested_image_size = _parse_size(_requested_image_size_node)

        for part_id, part in node["partitions"].items():
            label = part_id
            entry_obj, gpt_type, gpt_guid = self._process_entry(part)
            self._partitions.append(GPTPartition(label, gpt_type, gpt_guid, start=0, size=0, entry=entry_obj))

    def size(self) -> int:
        "Returns size of image in bytes. Requested in yaml or actually calculated."
        if not self._size:
            self._complete_init()

        if self._requested_image_size:
            if self._requested_image_size < self._size:
                raise Exception(
                    f"Actual size ({self._size}) of image is bigger than requested one ({self._requested_image_size}).")
            else:
                self._size = self._requested_image_size

        return self._size

    @staticmethod
    def _process_entry(node: YamlValue):
        entry_obj = construct_entry(node)
        gpt_type = node.get("gpt_type", "").as_str
        if not gpt_type:
            log.warning("No GPT type is provided %s, using default", node.mark)
            gpt_type = "8DA63339-0007-60C0-C436-083AC8230908"

        gpt_guid = node.get("gpt_guid", "").as_str

        return (entry_obj, gpt_type, gpt_guid)

    def _complete_init(self):
        partitions = [x._replace(size=x.entry.size()) for x in self._partitions]
        self._partitions, self._size = sfdisk.fixup_partition_table(partitions)

    def write(self, fp, offset):
        if not self._size:
            self._complete_init()
        if offset == 0:
            sfdisk.write(fp, self._partitions)
        else:
            # Write partition into temporary file, then copy it into
            # resulting file
            with NamedTemporaryFile("wb") as tempf:
                tempf.truncate(self._size)
                sfdisk.write(tempf, self._partitions)
                ext_utils.dd(tempf, fp, offset)

        for part in self._partitions:
            part.entry.write(fp, part.start + offset)

    def get_deps(self) -> List[str]:
        "Return list of dependencies needed to build this block"
        return list(itertools.chain().from_iterable(
            [part.entry.get_deps() for part in self._partitions]))


class RawImage(BlockEntry):
    "Represents raw image file which needs to be copied as is"

    def __init__(self, node: YamlValue):
        self._node = node
        self._fname = self._node["image_path"].as_str
        self._size = 0

    def _complete_init(self):
        mark = self._node["image_path"].mark
        if not os.path.exists(self._fname):
            raise YAMLProcessingError(f"Can't find file '{self._fname}'", mark)
        fsize = os.path.getsize(self._fname)
        size_node = self._node.get("size", None)
        if size_node:
            self._size = _parse_size(size_node)
            if fsize > self._size:
                raise YAMLProcessingError(
                    f"File '{self._fname}' is bigger than partition entry ({self._size})",
                    size_node.mark)
        else:
            self._size = fsize

    def size(self) -> int:
        "Returns size in bytes"
        if not self._size:
            self._complete_init()
        return self._size

    def write(self, fp, offset):
        if not self._size:
            self._complete_init()
        ext_utils.dd(self._fname, fp, offset)

    def get_deps(self) -> List[str]:
        "Return list of dependencies needed to build this block"
        return [self._fname]


class AndroidSparse(BlockEntry):
    "Represents android sparse image file"

    def __init__(self, node: YamlValue):
        self._node = node
        self._fname = self._node["image_path"].as_str
        self._size = 0

    def _read_size(self, mark: Mark):
        # pylint: disable=invalid-name
        FMT = "<IHHHHIIII"
        MAGIC = 0xed26ff3a
        size = struct.calcsize(FMT)
        with open(self._fname, "rb") as data:
            buf = data.read(size)
            if len(buf) < size:
                raise YAMLProcessingError(
                    f"Not enough data for Android sparse header in '{self._fname}'", mark)
            header = struct.unpack(FMT, buf)
            if header[0] != MAGIC:
                raise YAMLProcessingError(f"Invalid Android sparse header in '{self._fname}'", mark)
            # blk_sz * total_blks
            return header[5] * header[6]

    def _complete_init(self):
        mark = self._node["image_path"].mark
        if not os.path.exists(self._fname):
            raise YAMLProcessingError(f"Can't find file '{self._fname}'", mark)
        fsize = self._read_size(mark)
        size_node = self._node.get("size", None)
        if size_node:
            self._size = _parse_size(size_node)
            if fsize > self._size:
                raise YAMLProcessingError(
                    f"Un-sparesd file '{self._fname}' is bigger than partition entry",
                    size_node.mark)
        else:
            self._size = fsize

    def size(self) -> int:
        "Returns size in bytes"
        if not self._size:
            self._complete_init()
        return self._size

    def write(self, fp, offset):
        if not self._size:
            self._complete_init()
        with NamedTemporaryFile("w+b", dir=".") as tmpf:
            ext_utils.simg2img(self._fname, tmpf)
            ext_utils.dd(tmpf, fp, offset)

    def get_deps(self) -> List[str]:
        "Return list of dependencies needed to build this block"
        return [self._fname]


class EmptyEntry(BlockEntry):
    "Represents empty partition"

    def __init__(self, node: YamlValue):
        self._size = _parse_size(node["size"])
        self._fill_by_zero = (node.get("filled", "").as_str == "zeroes")

    def size(self) -> int:
        "Returns size in bytes"
        return self._size

    def write(self, fp, offset):
        if self._fill_by_zero:
            ext_utils.dd("/dev/zero", fp, offset, out_size=self._size)


class FileSystem(BlockEntry):
    "Represents a filesystem with list of files"

    def __init__(self, node: YamlValue):
        self._node = node
        self._size = 0
        self._items: List[Tuple[str, str, Mark]] = []

        files_node = self._node.get("files", None)
        if files_node:
            log.warn("Usage of 'files' is deprecated. Use 'items' please.")
            for remote_name, local_node in cast(YamlValue, files_node).items():
                self._items.append((remote_name, local_node.as_str, local_node.mark))

        items_node = self._node.get("items", None)
        if items_node:
            for remote_name, local_node in cast(YamlValue, items_node).items():
                self._items.append((remote_name, local_node.as_str, local_node.mark))

    def _complete_init(self):
        for _, local_name, local_mark in self._items:
            if not os.path.isfile(local_name) and not os.path.isdir(local_name):
                raise YAMLProcessingError(f"Can't find file '{local_name}'", local_mark)

        # calculate size of original files and directories
        files_size = 0
        for remote_path, local_path, _ in self._items:
            if os.path.isfile(local_path):
                files_size += os.path.getsize(local_path)
            if os.path.isdir(local_path):
                files_size += os.path.getsize(local_path)
                for (dirpath, dirnames, filenames) in os.walk(local_path, topdown=True):
                    for filename in filenames:
                        # we may have links to location that is incorrect on host, so we use lstat to handle this
                        files_size += os.lstat(os.path.join(dirpath, filename)).st_size
                    for dirname in dirnames:
                        files_size += os.path.getsize(os.path.join(dirpath, dirname))
        files_size += 8 * 1024 * 1024

        size_node = self._node.get("size", None)
        if size_node:
            self._size = _parse_size(size_node)
            if files_size > self._size:
                raise YAMLProcessingError(
                    f"Computed size is {files_size}, it is bigger than partition size {self._size}",
                    size_node.mark)
        else:
            self._size = files_size

    def size(self) -> int:
        "Returns size in bytes"
        if not self._size:
            self._complete_init()
        return self._size

    def write(self, fp, offset):
        raise NotImplementedError()

    def get_deps(self) -> List[str]:
        "Return list of dependencies needed to build this block"
        return [f[1] for f in self._items]


class Ext4(FileSystem):
    "Represents ext4 fs with list of files"
    def write(self, fp, offset):
        if not self._size:
            self._complete_init()
        with NamedTemporaryFile() as tempf, TemporaryDirectory() as tempd:
            for remote, local, _ in self._items:
                # user can specify destination folder from root
                # and we need to remove very first '/' for correct
                # work of os.path.join
                remote = remote.lstrip('/')
                # create destination subfolders
                remote_path_and_name = os.path.split(remote)
                if remote_path_and_name[0]:
                    # create destination subfolder
                    os.makedirs(os.path.join(tempd, remote_path_and_name[0]), exist_ok=True)

                if os.path.isfile(local):
                    shutil.copyfile(local, os.path.join(tempd, remote))
                if os.path.isdir(local):
                    shutil.copytree(local, os.path.join(tempd, remote), symlinks=True, dirs_exist_ok=True)
            tempf.truncate(self._size)
            ext_utils.mkext4fs(tempf, tempd)
            ext_utils.dd(tempf, fp, offset)


class Vfat(FileSystem):
    "Represents vfat fs with list of files"

    def unwrap_dirs(self):
        "Return list of files with flatten content of the directories"
        out_list: List[Tuple[str, str, Mark]] = []
        for remote, local, mark in self._items:
            if os.path.isfile(local):
                out_list.append([remote, local, mark])
            if os.path.isdir(local):
                for (dirpath, _, filenames) in os.walk(local, topdown=True):
                    remote_dirpath = dirpath.replace(local, remote, 1)
                    for filename in filenames:
                        # we skip symlinks as not supported on vfat
                        if os.path.islink(os.path.join(dirpath, filename)):
                            log.warn("Symlink '%s' is skipped.", os.path.join(dirpath, filename))
                        else:
                            out_list.append([os.path.join(remote_dirpath, filename),
                                             os.path.join(dirpath, filename), mark])
        return out_list

    def write(self, fp, offset):
        if not self._size:
            self._complete_init()
        with NamedTemporaryFile() as tempf:
            tempf.truncate(self._size)
            ext_utils.mkvfatfs(tempf)
            # for vfat we have to create each subdir (if any) and copy file one by one
            # that's why we need to 'unwrap' content of any input directory
            self._items = self.unwrap_dirs()
            # scan all remote filenames and collect the list of folders to create
            list_for_mmd = list()
            for remote, _, _ in self._items:
                # remove starting '/' to avoid:
                # - icluding different forms of same name, like "/zxc" and "zxc"
                # - adding root folder "/" to list
                remote_path_and_name = os.path.split(remote.lstrip('/'))
                if remote_path_and_name[0]:
                    # Here is explanation of what we do here.
                    # mmd can't create chain of folders like '/a/s/d' if parent
                    # folders (/a/s/) do not exist.
                    # You need to create each subfolder by separate call or provide them
                    # in incremental way, like `mmd /a /a/s /a/s/d`.
                    # Taking into account that we work with an image file,
                    # we need to prefix each folder with "::".
                    # So here we split each folder into slices,
                    # and re-assemble slices in incremental way.
                    current_path = ""
                    for path_slice in remote_path_and_name[0].split('/'):
                        if path_slice:
                            current_path += "/" + path_slice
                            list_for_mmd.append("::" + current_path)
            if list_for_mmd:
                # Remove duplicates by transforming the list of strings to the set of strings.
                # And sort strings because after 'set()' we may have 'a/s/d' before 'a/s' in some cases.
                list_for_mmd = sorted(set(list_for_mmd))
                # create all destination subfolders at once
                ext_utils.mmd(tempf, list_for_mmd)

            for remote, local, _ in self._items:
                ext_utils.mcopy(tempf, local, remote)
            ext_utils.dd(tempf, fp, offset)


_ENTRY_TYPES = {
    "gpt": GPT,
    "raw_image": RawImage,
    "ext4": Ext4,
    "vfat": Vfat,
    "empty": EmptyEntry,
    "android_sparse": AndroidSparse,
}


def construct_entry(node: YamlValue) -> BlockEntry:
    "Construct BlockEntry object from YAML node"
    entry_type = node["type"]
    if entry_type.as_str not in _ENTRY_TYPES:
        raise YAMLProcessingError(f"Unknown type '{entry_type.as_str}'", entry_type.mark)

    return _ENTRY_TYPES[entry_type.as_str](node)


_SUFFIXES = {
    "B": 1,
    "KB": 1000,
    "MB": 1000000,
    "GB": 1000000000,
    "TB": 1000000000000,
    "KiB": 1024,
    "MiB": 1024 * 1024,
    "GiB": 1024 * 1024 * 1024,
    "TiB": 1024 * 1024 * 1024 * 1024,
}


def _parse_size(node: YamlValue) -> int:
    components = node.as_str.split(" ")
    if len(components) == 1:
        return int(components[0])
    if len(components) == 2:
        suffix = components[1]
        if suffix not in _SUFFIXES:
            raise YAMLProcessingError(f"Unknown size suffix '{suffix}'", node.mark)
        scaler = _SUFFIXES[suffix]
        return int(components[0]) * scaler
    raise YAMLProcessingError(f"Can't parse size entry '{node.as_str}'", node.mark)



#!/usr/bin/env python3

# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
Console entry point for moulin
"""

from moulin.main import moulin_entry


def main():
    moulin_entry()


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 EPAM Systems
"""
Zephyr build generator
"""

import os.path
from typing import List
from moulin.yaml_wrapper import YamlValue
from moulin import ninja_syntax
from moulin.utils import construct_fetcher_dep_cmd


def get_builder(conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                generator: ninja_syntax.Writer):
    """
    Return configured Zephyr builder
    """
    return ZephyrBuilder(conf, name, build_dir, src_stamps, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """
    Generate Zephyr build rules for ninja
    """
    cmd = " && ".join([
        # Generate fetcher dependency file
        construct_fetcher_dep_cmd(),
        "cd $build_dir",
        "source zephyr/zephyr-env.sh",
        "$env west build -p auto -b $board -d $work_dir $target $shields",
    ])
    generator.rule("zephyr_build",
                   command=f'bash -c "{cmd}"',
                   description="Invoke Zephyr build system",
                   pool="console",
                   deps="gcc",
                   depfile=".moulin_$name.d",
                   restat=True)
    generator.newline()


class ZephyrBuilder:
    """
    ZephyrBuilder class generates Ninja rules for Zephyr build configuration
    """
    def __init__(self, conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                 generator: ninja_syntax.Writer):
        self.conf = conf
        self.name = name
        self.generator = generator
        self.src_stamps = src_stamps
        self.build_dir = build_dir

    def gen_build(self):
        """Generate Ninja rules to build Zephyr"""

        env_node = self.conf.get("env", None)
        if env_node:
            env_values = [x.as_str for x in env_node]
        else:
            env_values = []
        env = " ".join(env_values)

        shields_node = self.conf.get("shields", None)
        if shields_node:
            shields_vals = [x.as_str for x in shields_node]
            shields = f'-DSHIELD=\\"{" ".join(shields_vals)}\\"'
        else:
            shields = ""

        variables = {
            "name": self.name,
            "build_dir": self.build_dir,
            "board": self.conf["board"].as_str,
            "target": self.conf["target"].as_str,
            "work_dir": self.conf.get("work_dir", "zephyr/build").as_str,
            "shields": shields,
            "env": env,
        }
        targets = self.get_targets()
        deps = list(self.src_stamps)

        additional_deps_node = self.conf.get("additional_deps", None)

        if additional_deps_node:
            deps.extend([d.as_str for d in additional_deps_node])

        self.generator.build(targets, "zephyr_build", deps, variables=variables)
        self.generator.newline()

        return targets

    def get_targets(self):
        "Return list of targets that are generated by this build"
        return [os.path.join(self.build_dir, t.as_str) for t in self.conf["target_images"]]

    def capture_state(self):
        """
        This method should capture state for reproducible builds.
        """



"""
Intergration-level test for build generator
"""

import unittest
from unittest.mock import patch, ANY

import yaml

from moulin.build_generator import generate_build
from moulin.build_conf import MoulinConfiguration


class TestBuilder(unittest.TestCase):
    def setUp(self):
        patcher = patch('moulin.ninja_syntax.Writer')
        self.Writer = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = patch('moulin.build_generator.open')
        self.opener = patcher.start()
        self.addCleanup(patcher.stop)

    def test_minimal(self):
        doc = """
desc: "Test build"
components:
  test:
    builder:
      type: "null"
    sources:
      - type: "null"
        """
        node = yaml.compose(doc)
        conf = MoulinConfiguration(node)
        generate_build(conf, "test.yaml")
        self.Writer.return_value.build.assert_any_call("fetch-test", "phony", ANY)
        self.Writer.return_value.build.assert_any_call("test", "phony", ANY)
        self.Writer.return_value.build.assert_any_call("build.ninja", "regenerate", ANY)

    def test_set_default(self):
        doc = """
desc: "Test build"
components:
  test:
    sources:
      - type: "null"
    default: true
    builder:
      type: "null"
        """
        node = yaml.compose(doc)
        conf = MoulinConfiguration(node)
        generate_build(conf, "test.yaml")
        self.Writer.return_value.default.assert_called_with("test")

    def test_no_default(self):
        doc = """
desc: "Test build"
components:
  test:
    sources:
      - type: "null"
    builder:
      type: "null"
        """
        node = yaml.compose(doc)
        conf = MoulinConfiguration(node)
        generate_build(conf, "test.yaml")
        self.Writer.return_value.default.assert_not_called()

    def test_flatten_sources(self):
        doc = """
desc: "Test build"
components:
  test:
    builder:
      type: "null"
    sources:
      -
        - type: "null"
        """
        node = yaml.compose(doc)
        conf = MoulinConfiguration(node)
        generate_build(conf, "test.yaml")



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""Git fetcher module"""

import os.path
from typing import List
import pygit2
from moulin.yaml_wrapper import YamlValue
from moulin.utils import create_stamp_name
from moulin import ninja_syntax


def get_fetcher(conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
    """Construct and return GitFetcher object"""
    return GitFetcher(conf, build_dir, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """Generate build rules using Ninja generator"""
    # If $git_url uses ssh protocol and is not in .ssh/known_hosts,
    # the ssh client starts to iterate with a user.
    # But this interaction is not visible and
    # looks just like the ninja hung on some operation.
    # So we have to use ssh in BatchMode to throw
    # an error instead of asking invisible questions.
    generator.rule("git_clone",
                   command="GIT_SSH_COMMAND='ssh -o BatchMode=yes' "
                           "git clone -q $git_url $git_dir && touch $out",
                   description="git clone")
    generator.newline()

    generator.rule("git_checkout",
                   command="git -C $git_dir checkout -q $git_rev && touch $out",
                   description="git checkout")
    generator.newline()


def _guess_dirname(url: str):
    # TODO: Add support for corner cases
    if url.endswith(".git"):
        url = url[:-4]
    if url.endswith("/"):
        url = url[:-1]
    return url.split("/")[-1]


_SEEN_REPOS = []


class GitFetcher:
    """Git fetcher class. Provides methods to generate rules for fetching git repositories"""
    def __init__(self, conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
        self.conf = conf
        self.build_dir = build_dir
        self.generator = generator
        self.url = conf["url"].as_str
        dirname = conf.get("dir", default=_guess_dirname(self.url)).as_str
        self.git_dir = os.path.join(build_dir, dirname)
        self.git_rev = conf.get("rev", default="master").as_str

    def gen_fetch(self):
        """Generate instruction to fetch git repo"""
        clone_target = self.git_dir
        checkout_stamp = create_stamp_name(self.build_dir, self.url, "checkout")

        # Do not checkout repos for the second time
        if checkout_stamp in _SEEN_REPOS:
            return checkout_stamp

        _SEEN_REPOS.append(checkout_stamp)
        self.generator.build(clone_target,
                             "git_clone",
                             variables={
                                 "git_url": self.url,
                                 "git_dir": self.git_dir
                             })
        self.generator.newline()
        self.generator.build(checkout_stamp,
                             "git_checkout",
                             clone_target,
                             variables={
                                 "git_rev": self.git_rev,
                                 "git_dir": self.git_dir
                             })
        self.generator.newline()
        return checkout_stamp

    def get_file_list(self) -> List[str]:
        "Get list of files under git control"
        repo = pygit2.Repository(self.git_dir)
        index = repo.index
        index.read()
        return [os.path.join(self.git_dir, entry.path) for entry in index]

    def capture_state(self):
        """
        Update provided conf with the actual commit ID. This is
        mostly needed for build history/reproducible builds.
        """
        repo = pygit2.Repository(self.git_dir)
        head = repo.revparse_single("HEAD")
        self.conf["rev"] = str(head)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""Repo fetcher module"""

import shlex
import os.path
import subprocess
from typing import List
import pygit2
from moulin.utils import create_stamp_name
from moulin.yaml_wrapper import YamlValue
from moulin import ninja_syntax


def get_fetcher(conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
    """Construct and return RepoFetcher object"""
    return RepoFetcher(conf, build_dir, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """Generate build rules using Ninja generator"""
    cmd = " && ".join([
        "mkdir -p $repo_dir",
        "cd $repo_dir",
        "repo init -u $url $repo_args",
    ])
    generator.rule("repo_init", command=cmd, description="Initialize repo directory")
    generator.newline()

    generator.rule("repo_sync",
                   command="cd $repo_dir && repo sync && touch $out",
                   description="Repo sync")
    generator.newline()


class RepoFetcher:
    """
    Repo fetcher class. Provides methods to generate rules for
    fetching repo-based repositories
    """
    def __init__(self, conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
        self.conf = conf
        self.build_dir = build_dir
        self.generator = generator
        self.url = conf["url"].as_str
        dirname = conf.get("dir", default=".").as_str
        self.repo_dir = os.path.join(build_dir, dirname)

    def gen_fetch(self):
        """Generate instructions to fetch repo-based repository"""
        repo_args = []
        manifest = self.conf.get("manifest", "").as_str
        if manifest:
            repo_args.append(f"-m {shlex.quote(manifest)}")
        rev = self.conf.get("rev", "").as_str
        if rev:
            repo_args.append(f"-b {shlex.quote(rev)}")
        depth = self.conf.get("depth", 0).as_int
        if depth:
            repo_args.append(f"--depth={depth}")
        groups = self.conf.get("groups", "").as_str
        if groups:
            repo_args.append(f"-g {groups}")

        init_target = os.path.join(self.repo_dir, ".repo")
        sync_stamp = create_stamp_name(self.build_dir, self.url, "sync")

        self.generator.build(init_target,
                             "repo_init",
                             variables={
                                 "repo_dir": self.repo_dir,
                                 "url": self.url,
                                 "repo_args": " ".join(repo_args)
                             })
        self.generator.newline()

        self.generator.build(sync_stamp,
                             "repo_sync",
                             init_target,
                             variables={"repo_dir": self.repo_dir})
        self.generator.newline()

        return sync_stamp

    def get_file_list(self) -> List[str]:
        "Get list of files under repo control"

        # First, get list of projects
        repo_out = subprocess.run(["repo", "list"],
                                  check=True,
                                  cwd=self.repo_dir,
                                  stdout=subprocess.PIPE,
                                  encoding="utf-8")
        projects = [x.split(":")[0].strip() for x in repo_out.stdout.split("\n")]

        # Then interrogate each git repo
        result: List[str] = []
        for project in projects:
            if len(project) == 0:
                continue
            git_dir = os.path.join(self.repo_dir, project)
            repo = pygit2.Repository(git_dir)
            index = repo.index
            index.read()
            for entry in index:
                path = os.path.join(git_dir, entry.path)
                if os.path.isfile(path):
                    result.append(path.replace("$", "$$"))
        return result

    def capture_state(self):
        """
        Update provided conf with the current repo state. This is
        mostly needed for build history/reproducible builds.
        """
        raise Exception("capture_state is not implemented for repo fetcher")



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
This module takes processed YAML config tree (with variables expanded
and parameters applied) and produces Ninja build file
"""

import os.path
import sys
from typing import Optional, List

from importlib import import_module
from moulin import ninja_syntax
from moulin import make_syntax
from moulin import rouge
from moulin import yaml_helpers as yh
from moulin.build_conf import MoulinConfiguration

BUILD_FILENAME = 'build.ninja'


def generate_build(conf: MoulinConfiguration,
                   conf_file_name,
                   ninja_build_fname=BUILD_FILENAME) -> None:
    """
    Write Ninja build file based on pre-processed config tree
    """
    generator = ninja_syntax.Writer(open(ninja_build_fname, 'w'), width=120)

    _gen_regenerate(conf_file_name, generator)

    rouge.gen_build_rules(generator)

    _flatten_sources(conf)
    # We want to have all Ninja build rules before all actual build
    # commands. So we need to scan conf twice. On the first scan we will
    # determine and load all required plugins. On the same time, we'll ask them
    # to generate Ninja rules.
    builder_modules, fetcher_modules = _get_modules(conf, generator)

    root = conf.get_root()
    # Now we have all plugins loaded and we can generate some build rules
    for comp_name, component in root["components"].items():
        build_dir = component.get("build-dir", comp_name).as_str
        builder_conf = component["builder"]

        source_stamps = []
        if "sources" in component:
            for source in component["sources"]:
                source_type = source["type"].as_str
                fetcher_module = fetcher_modules[source_type]
                fetcher = fetcher_module.get_fetcher(source, build_dir, generator)
                fetcher_stamps = fetcher.gen_fetch()
                if isinstance(fetcher_stamps, list):
                    source_stamps.extend(fetcher_stamps)
                else:
                    source_stamps.append(fetcher_stamps)

        # Generate handy 'fetch-{component}' rule
        generator.build(f"fetch-{comp_name}", "phony", source_stamps)
        generator.newline()

        builder_type = builder_conf["type"].as_str
        builder_module = builder_modules[builder_type]
        builder = builder_module.get_builder(builder_conf, comp_name, build_dir, source_stamps,
                                             generator)

        build_stamps = builder.gen_build()
        generator.build(comp_name, "phony", build_stamps)
        generator.newline()
        if component.get("default", False).as_bool:
            generator.default(comp_name)

    rouge.gen_build(generator, rouge.get_available_images(conf.get_root()))


def generate_fetcher_dyndep(conf: MoulinConfiguration, component: str):
    _flatten_sources(conf)
    generator = make_syntax.Writer(open(f".moulin_{component}.d", 'w'), width=120)

    builder_modules, fetcher_modules = _get_modules(conf, None)
    component_node = conf.get_root()["components"][component]
    build_dir = component_node.get("build-dir", component).as_str
    builder_node = component_node["builder"]
    builder_type = builder_node["type"].as_str
    builder_module = builder_modules[builder_type]
    builder = builder_module.get_builder(builder_node, component, build_dir, [], generator)

    deps: List[str] = []
    targets = builder.get_targets()
    if "sources" in component_node:
        for source in component_node["sources"]:
            source_type = source["type"].as_str
            fetcher_module = fetcher_modules[source_type]
            fetcher = fetcher_module.get_fetcher(source, build_dir, generator)
            deps.extend(fetcher.get_file_list())
    generator.simple_dep(targets, deps)


def _gen_regenerate(conf_file_name, generator: ninja_syntax.Writer):
    this_script = os.path.abspath(sys.argv[0])
    args = " ".join(sys.argv[1:])
    generator.rule("regenerate", command=f"{this_script} {args}", generator=1)
    generator.newline()
    generator.build(BUILD_FILENAME, "regenerate", [this_script, conf_file_name])
    generator.newline()


def _flatten_sources(conf: MoulinConfiguration):
    for _, component in yh.get_mandatory_mapping(conf.get_root_node(), "components"):
        if yh.get_node(component, "sources"):
            yh.flatten_list(yh.get_mandatory_sequence_node(component, "sources"))


def _get_modules(conf: MoulinConfiguration, generator: Optional[ninja_syntax.Writer]):
    builder_modules = {}
    fetcher_modules = {}
    for _, component in conf.get_root()["components"].items():
        b_type = component["builder"]["type"].as_str
        if b_type not in builder_modules:
            builder_modules[b_type] = _prepare_builder(b_type, generator)
        if "sources" in component:
            for source in component["sources"]:
                f_type = source["type"].as_str
                if f_type not in fetcher_modules:
                    fetcher_modules[f_type] = _prepare_fetcher(f_type, generator)
    return builder_modules, fetcher_modules


def _prepare_builder(builder, generator):
    module = import_module(f".builders.{builder}", __package__)
    if generator:
        module.gen_build_rules(generator)
    return module


def _prepare_fetcher(fetcher, generator):
    module = import_module(f".fetchers.{fetcher}", __package__)
    if generator:
        module.gen_build_rules(generator)
    return module



# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 EPAM Systems
"""
Archive packer builder module
"""

import os.path
from typing import List
from moulin.yaml_wrapper import YamlValue
from moulin import ninja_syntax


def get_builder(conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                generator: ninja_syntax.Writer):
    """
    Return configured ArchiveBuilder class
    """
    return ArchiveBuilder(conf, name, build_dir, src_stamps, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """
    Generate yocto build rules for ninja
    """

    generator.rule("tar_pack",
                   "tar --sparse --dereference --create --auto-compress --file=$out $items",
                   description="Archiving into $out",
                   restat=True)
    generator.newline()


class ArchiveBuilder:
    """
    ArchiveBuilder class generates Ninja rules to archive files
    """

    def __init__(self, conf: YamlValue, name: str, build_dir: str, src_stamps: List[str],
                 generator: ninja_syntax.Writer):
        self.conf = conf
        self.name = name
        self.generator = generator
        self.src_stamps = src_stamps
        self.build_dir = build_dir

    def gen_build(self):
        """Generate ninja rules to create archive"""
        items = [x.as_str for x in self.conf["items"]]

        targets = self.get_targets()
        deps = list(self.src_stamps)
        deps.extend(items)

        self.generator.build(targets, "tar_pack", deps, variables={"items": items})
        self.generator.newline()

        return targets

    def get_targets(self):
        "Return list of targets that are generated by this build"
        return [os.path.join(self.build_dir, self.conf["name"].as_str)]

    def capture_state(self):
        """
        This method should capture state for reproducible builds.
        Luckily, there is nothing to do, as this builder just arhives
        existing files
        """



from setuptools import setup

from typing import Dict, Any

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

SETUP_ARGS: Dict[str, Any] = dict(
    name='moulin',  # Required
    version='0.15',  # Required
    description='Meta-build system',  # Required
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/xen-troops/moulin',  # Optional
    author='Volodymyr Babchuk',  # Optional
    author_email='volodymyr_babchuk@epam.com',  # Optional

    # Note that this is a string of words separated by whitespace, not a list.
    keywords='build ninja yocto android repo git',
    packages=[
        "moulin",
        "moulin.fetchers",
        "moulin.builders",
        "moulin.rouge",
    ],  # Required
    install_requires=[
        'pygit2',
        'importlib_metadata',
        'packaging',
        'pyaml',
    ],
    python_requires=">=3.6",
    entry_points={
        'console_scripts': ['moulin = moulin.main:moulin_entry', 'rouge = moulin.main:rouge_entry']
    })
setup(**SETUP_ARGS)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 EPAM Systems
"""Null fetcher module"""

from typing import List

from moulin.yaml_wrapper import YamlValue
from moulin import ninja_syntax


def get_fetcher(conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
    """Construct and return NullFetcher object"""
    return NullFetcher(conf, build_dir, generator)


def gen_build_rules(generator: ninja_syntax.Writer):
    """Generate build rules using Ninja generator"""
    generator.newline()


class NullFetcher:
    """
    Repo fetcher class. Provides methods to generate rules for
    fetching repo-based repositories
    """
    def __init__(self, conf: YamlValue, build_dir: str, generator: ninja_syntax.Writer):
        pass

    def gen_fetch(self):
        """Generate instructions to fetch repo-based repository"""
        return "null.stamp"

    def get_file_list(self) -> List[str]:
        """Get list of files under fetcher control"""
        return []

    def capture_state(self):
        """Update provided conf with the current fetcher state"""



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 EPAM Systems
"""
Rouge Image generator
"""
import sys
from typing import List, NamedTuple

from moulin import ninja_syntax
from moulin.yaml_wrapper import YamlValue
from .block_entry import construct_entry


class RougeImage(NamedTuple):
    "Represents rouge image"
    name: str
    desc: str
    node: YamlValue


def get_available_images(root_node: YamlValue) -> List[RougeImage]:
    "Return list of available images from YAML config"
    images_node = root_node.get("images", None)
    if not images_node:
        return []

    ret: List[RougeImage] = []
    for name, image_node in images_node.items():
        desc = image_node.get("desc", "No description available").as_str
        ret.append(RougeImage(name, desc, image_node))
    return ret


def gen_build_rules(generator: ninja_syntax.Writer):
    "Generate ninja rules to build images via rouge"
    moulin_args = " ".join(sys.argv[1:])
    generator.rule("rouge", f"rouge {moulin_args} -fi $image_name -o $out")
    generator.newline()
    generator.rule("gzip", "gzip -1kf $in")
    generator.newline()
    generator.rule("bmaptool", "bmaptool create $in -o $out")
    generator.newline()


def gen_build(generator: ninja_syntax.Writer, images: List[RougeImage]):
    "Generate build rules for given list of images"
    for image in images:
        block_entry = construct_entry(image.node)
        generator.build(f"{image.name}.img",
                        "rouge",
                        block_entry.get_deps(),
                        variables=dict(image_name=image.name),
                        pool="console")
        generator.build(f"{image.name}.img.gz",
                        "gzip",
                        f"{image.name}.img",
                        pool="console")
        generator.build(f"{image.name}.img.bmap",
                        "bmaptool",
                        f"{image.name}.img",
                        pool="console")
        generator.build(f"image-{image.name}", "phony", f"{image.name}.img")
